"""
---
title: Graph Neural Networks
summary: >
 A set of PyTorch implementations/tutorials related to graph neural networks
---

# Graph Neural Networks

* [Graph Attention Networks (GAT)](gat/index.html)
* [Graph Attention Networks v2 (GATv2)](gatv2/index.html)
"""
